/*
 * Copyright (C) 2008 Gauvain Pocentek <gpocentek@linutop.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>

#include <gmodule.h>
#include <glib/gprintf.h>
#include <gtk/gtk.h>
#include <libxfce4util/libxfce4util.h>
#include <libxfcegui4/libxfcegui4.h>

#include "appstarter.h"

#define MARGIN 2

static const char *dentry_keys [] = {
    "Name",
    "GenericName",
    "Exec",
    "Icon",
};

/* Global variables */
interface *ui;
int compt = 0;
app *active_item;

/* Read the data dir and build a list of available plugins */
static GList *
build_modules_list ()
{
    GDir   *dirp;
    const gchar  *name;
    GList  *list = NULL;

    gchar   *fullpath = NULL;
    gchar   *sopath = NULL;
    gchar   *value = NULL;
    XfceDesktopEntry *de;

    dirp = g_dir_open (DESKTOPDIR, 0, NULL);
    if (G_UNLIKELY(! dirp))
        return list;

    /* The None plugin first */
    app *newapp;
    newapp = g_new0(app, 1);
    newapp->display_name = g_strdup (_("None"));
    newapp->name = g_strdup ("none");
    newapp->icon = g_strdup ("/usr/share/icons/Rodent/48x48/stock/stock_cancel.png");
    sopath = LIBDIR "/appstarter/plugins/none-plugin.so";
    newapp->module = g_module_open (sopath, G_MODULE_BIND_LOCAL);
    if (newapp->module)
        list = g_list_append (list, newapp);

    /* plugins from .desktop files */
    while ((name = g_dir_read_name (dirp)) != NULL)
    {
        fullpath = g_strconcat(DESKTOPDIR, "/", name, NULL);
        de = xfce_desktop_entry_new (fullpath, dentry_keys,
                                     G_N_ELEMENTS (dentry_keys));
        g_free (fullpath);

        if (G_UNLIKELY (!de))
            continue;
        newapp = g_new0(app, 1);

        xfce_desktop_entry_get_string (de, "GenericName", TRUE, &value);
        if (value)
            newapp->display_name = g_strdup (value);
        else
            newapp->display_name = g_strdup ("");
        xfce_desktop_entry_get_string (de, "Name", TRUE, &value);
        if (value)
            newapp->name = g_strdup (value);
        else
            newapp->name = g_strdup ("");
        xfce_desktop_entry_get_string (de, "Icon", TRUE, &value);
        if (value)
            newapp->icon = g_strdup (value);
        else
            newapp->icon = g_strdup ("");
        xfce_desktop_entry_get_string (de, "Exec", TRUE, &value);
        if (value)
            sopath = g_strdup (value);
        else
            sopath = NULL;

        if (de)
            g_object_unref (de);

        if (!sopath)
            continue;

        newapp->module = g_module_open (sopath, G_MODULE_BIND_LOCAL);
        g_free (sopath);
        if (G_LIKELY (newapp->module))
            list = g_list_append (list, newapp);
        else
        {
            g_printf("Module import error: %s\n", g_module_error());
            g_free (newapp);
        }
        g_free (value);
    }

    g_dir_close(dirp);
    return list;
}

/* read the combo box selected item and return the associated
 * module
 */
app *
get_active_plugin ()
{
    GtkTreeModel  *model;
    GtkTreeIter    iter;
    app *item;

    model = gtk_combo_box_get_model(GTK_COMBO_BOX (ui->combo));
    gtk_combo_box_get_active_iter (GTK_COMBO_BOX (ui->combo), &iter);
    gtk_tree_model_get (model, &iter, 2, &item, -1);

    return item;
}

/* combo changed callback */
static void
plugin_changed (GtkWidget *combo, gpointer data)
{
    GetWidgetFunc  get_widget;
    FreeFunc       free_plugin;
    app           *item;

    if (active_item)
    {
        item = active_item;
        g_module_symbol (item->module, "free_all",
                (gpointer *) &free_plugin);
        gtk_widget_destroy (ui->config);
        free_plugin();
    }

    item = get_active_plugin ();
    active_item = item;

    g_module_symbol (item->module, "get_config_widget",
            (gpointer *) &get_widget);
    ui->config = get_widget(ui->conffile);
    if (strcmp (item->icon, ""))
    {
        GdkPixbuf *pix = gdk_pixbuf_new_from_file_at_size (item->icon, 36, 36, NULL);
        xfce_heading_set_icon (ui->head, pix ? pix : NULL);
    } else {
        xfce_heading_set_icon_name (ui->head, NULL);
        xfce_heading_set_icon (ui->head, NULL);
    }
    xfce_heading_set_title (ui->head, item->display_name);
    gtk_box_pack_start (GTK_BOX (ui->vbox), ui->config, FALSE, FALSE, MARGIN);
    gtk_widget_show (ui->config);
}

/* test btn clicked */
static void
test_btn_clicked (GtkWidget *w, gpointer data)
{
    app *item;
    GetCmdFunc get_cmd;
    gchar *tmp;

    item = get_active_plugin();
    if (strncmp (item->name, "none", 4) == 0)
        return;

    g_module_symbol (item->module, "get_cmd",
            (gpointer *) &get_cmd);
    tmp = g_strdup (get_cmd());
    if (tmp)
    {
        xfce_exec (tmp, FALSE, TRUE, NULL);
        g_free (tmp);
    }
}

/* build the dialog */
static void
build_ui()
{
    GtkWidget *hbox, *label, *hs;
    GtkCellRenderer *renderer;
    GtkWidget *test_btn;

    ui = g_new0(interface, 1);

    ui->dialog = gtk_dialog_new_with_buttons("AppStarter",
                                         NULL,
                                         GTK_DIALOG_MODAL,
                                         GTK_STOCK_OK,
                                         GTK_RESPONSE_ACCEPT,
                                         NULL);
    gtk_container_set_border_width (GTK_CONTAINER (ui->dialog), 5);   

    ui->vbox = gtk_vbox_new (FALSE, MARGIN);
    gtk_box_pack_start (GTK_BOX (GTK_DIALOG (ui->dialog)->vbox),
        ui->vbox, FALSE, FALSE, MARGIN);

    hbox = gtk_hbox_new (FALSE, MARGIN);
    gtk_box_pack_start (GTK_BOX (ui->vbox), hbox, FALSE, FALSE, MARGIN);

    label = gtk_label_new (_("Select an item: "));
    gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, MARGIN);

    ui->store = gtk_list_store_new(3, G_TYPE_INT, G_TYPE_STRING, G_TYPE_POINTER);
    ui->combo = gtk_combo_box_new_with_model (GTK_TREE_MODEL(ui->store));
    renderer = gtk_cell_renderer_text_new ();
    gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (ui->combo),
        renderer, TRUE);
    gtk_cell_layout_add_attribute (GTK_CELL_LAYOUT (ui->combo),
        renderer, "text", 1);
    gtk_box_pack_start (GTK_BOX (hbox), ui->combo, FALSE, FALSE, MARGIN);

    test_btn = gtk_button_new_with_label (_("Run a test"));
    gtk_box_pack_end (GTK_BOX (hbox), test_btn, FALSE, FALSE, 2);
    gtk_widget_show (test_btn);

    ui->onlyme = gtk_check_button_new_with_label (
        _("Start only this application (not the full desktop)"));
    gtk_box_pack_start (GTK_BOX (ui->vbox), ui->onlyme, FALSE, FALSE, MARGIN);

    hs = gtk_hseparator_new ();
    gtk_box_pack_start (GTK_BOX (ui->vbox), hs, TRUE, TRUE, MARGIN);

    ui->head = XFCE_HEADING (xfce_heading_new ());
    gtk_box_pack_start (GTK_BOX (ui->vbox), GTK_WIDGET (ui->head),
        TRUE, TRUE, MARGIN);

    ui->config = gtk_label_new(_("Please select an item."));
    gtk_box_pack_start (GTK_BOX (ui->vbox), ui->config, TRUE, TRUE, MARGIN);

    g_signal_connect(G_OBJECT (ui->combo), "changed",
        G_CALLBACK (plugin_changed), NULL);
    g_signal_connect(G_OBJECT (test_btn), "clicked",
        G_CALLBACK (test_btn_clicked), NULL);
}

/* populate the combo box (store) using the plugins GList */
static void
append_to_store (gpointer data, gpointer user_data)
{
    GtkTreeIter iter;

    app *item = (app *) data;
    gtk_list_store_append (ui->store, &iter);
    gtk_list_store_set (ui->store,
        &iter,
        0, compt,
        1, item->display_name,
        2, (gpointer *)item,
        -1);
    if (strcmp (ui->selected, item->name) == 0)
        gtk_combo_box_set_active (GTK_COMBO_BOX (ui->combo), compt);

    compt++;
}

static void
populate_store(GList *modules)
{
    g_list_foreach (modules, append_to_store, NULL);
}

/* config handlers */
static void
read_config()
{
    XfceRc *rc = as_rc_open (ui->conffile, TRUE);
    g_free (ui->selected);

    if (!rc)
    {
        ui->selected = g_strdup ("none");
        return;
    }

    xfce_rc_set_group(rc, "default");
    ui->selected = g_strdup (xfce_rc_read_entry (rc, "selected", "none"));
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (ui->onlyme),
        xfce_rc_read_bool_entry (rc, "nodesktop", FALSE));

    xfce_rc_close(rc);
}

static void
save_config()
{
    XfceRc *rc;
    GetCmdFunc get_cmd;
    app *item;

    rc = as_rc_open (ui->conffile, FALSE);
    if (G_UNLIKELY (!rc))
    {
        g_printf("Unable to save the configuration");
        return;
    }
    xfce_rc_set_group (rc, "default");

    item = get_active_plugin();
    xfce_rc_write_entry(rc, "selected", item->name);
    xfce_rc_write_bool_entry (rc, "nodesktop",
        gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (ui->onlyme)));

    g_module_symbol (item->module, "get_cmd",
            (gpointer *) &get_cmd);
    xfce_rc_write_entry (rc, "command", get_cmd());

    xfce_rc_close(rc);
}

gboolean
free_store (GtkTreeModel *model,
            GtkTreePath *path,
            GtkTreeIter *iter,
            gpointer data)
{
    app *item;
    gtk_tree_model_get (model, iter, 2, &item, -1);
    g_free (item->name);
    g_free (item->icon);
    g_free (item->display_name);
    g_module_close (item->module);
    g_free (item);
}

static void
interface_free ()
{
    gtk_tree_model_foreach (GTK_TREE_MODEL (ui->store),
                            (GtkTreeModelForeachFunc) free_store,
                            NULL);
    g_free (ui->selected);
    g_free (ui->conffile);
    g_free (ui);
}

int main (int argc, char *argv[])
{
    GList *modules;
    gchar *conffile = NULL;

    GOptionEntry options[] = {
        { "conffile", 'c', 0, G_OPTION_ARG_STRING,
          &conffile, N_("Use the specified configuration file"), "FILE" },
        { NULL },
    };

    xfce_textdomain (GETTEXT_PACKAGE, PACKAGE_LOCALE_DIR, "UTF-8");
    if (!gtk_init_with_args (&argc, &argv, "", options, PACKAGE, NULL)) {
        g_print ( _(
            "%s:\nTry %s --help to see a full list of available command line options.\n"),
                PACKAGE, PACKAGE_NAME);
        return 1;
    }

    build_ui();

    if (conffile)
        ui->conffile = g_strdup (conffile);
    else
        ui->conffile = NULL;
    g_free (conffile);

    modules = build_modules_list();
    read_config();
    populate_store(modules);

    gtk_widget_show_all(ui->dialog);
    int response = gtk_dialog_run (GTK_DIALOG (ui->dialog));
    if (response)
        save_config();

    interface_free ();

    return 0;
}
